#region Copyright Syncfusion Inc. 2001-2016.
// Copyright Syncfusion Inc. 2001-2016. All rights reserved.
// Use of this code is subject to the terms of our license.
// A copy of the current license can be obtained at any time by e-mailing
// licensing@syncfusion.com. Any infringement will be prosecuted under
// applicable laws. 
#endregion
using System;
using System.IO;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;

namespace WebHDFS
{
    class Program
    {
        static string userName,password,exitOption;
        static void Main(string[] args)
        {
            try
            {
                int clusterOption;
                bool clusterChoice = true;
                var clusterType=ClusterType.Normal;
                var commonHome = Directory.GetParent("..\\..\\..\\..\\..\\..\\..\\..\\").ToString();
                do
                {
                    Console.Write("Enter Namenode IP : ");
                    var hostname = Console.ReadLine();
                    Console.WriteLine("Please choose the cluster type (For localhost please choose normal cluster)\n1.Normal Cluster\n2.Secure Cluster\n3.Azure Cluster");
                    clusterOption = int.Parse(Console.ReadLine());
                    if (clusterOption >= 1 && clusterOption <= 3)
                    {
                        switch (clusterOption)
                        {
                            case 1:
                                clusterType = ClusterType.Normal;
                                break;
                            case 2:
                                clusterType = ClusterType.Secure;
                                Console.WriteLine(
                                    "Please enter the UserName(username@domainname) and Password\nUserName:");
                                userName = Console.ReadLine();
                                Console.WriteLine("password:");
                                password = Console.ReadLine();
                                ServicePointManager.ServerCertificateValidationCallback += ValidateServerCertificate;
                                hostname = GetHostName(hostname);
                                break;
                            case 3:
                                clusterType = ClusterType.Azure;
                                Console.WriteLine("Please enter the UserName and Password\nUserName:");
                                userName = Console.ReadLine();
                                Console.WriteLine("password:");
                                password = Console.ReadLine();
                                break;
                        }
                        int i;
                        do
                        {
                            Console.WriteLine("Enter any operation");
                            Console.WriteLine("1. Upload");
                            Console.WriteLine("2. Delete");
                            Console.WriteLine("3. MoveFile");
                            Console.WriteLine("4. MakeDirectory");
                            Console.WriteLine("5. SetReplication");
                            Console.WriteLine("6. Offset");
                            Console.WriteLine("7. Exit");
                            i = int.Parse(Console.ReadLine());
                            var p = new Program();
                            switch (i)
                            {
                                case 1:
                                    Console.WriteLine("Uploading file from '" + commonHome +
                                                      "\\BigDataSDK\\Samples\\Data\\WarPeace.txt'" +
                                                      " to hdfs location " + "'/Data/PracticeTemp.txt'");
                                    p.UploadFile(hostname, clusterType,
                                        commonHome + "\\BigDataSDK\\Samples\\Data\\WarPeace.txt",
                                        "/Data/PracticeTemp.txt");
                                    break;
                                case 2:
                                    Console.WriteLine("Deleting file from hdfs location " + "'/Data/PracticeTemp.txt'");
                                    if (p.DeleteFile(hostname, clusterType, "/Data/PracticeTemp.txt"))
                                        Console.WriteLine("Delete operation completed");
                                    else
                                        Console.WriteLine("Delete operation Failed");
                                    break;
                                case 3:
                                    Console.WriteLine("Move file from hdfs location " +
                                                      "'/Data/PracticeTemp.txt' to hdfs location " +
                                                      "'/Data/PracticeTemp2.txt'");
                                    if (p.MoveFile(hostname, clusterType, "/Data/PracticeTemp.txt",
                                        "/Data/PracticeTemp2.txt"))
                                        Console.WriteLine("Move operation completed");
                                    else
                                        Console.WriteLine("Move operation Failed");
                                    break;
                                case 4:
                                    Console.WriteLine("Make directory from hdfs location " + "'/Data/PracticeDirectory'");
                                    if (p.CreateDirectory(hostname, clusterType, "/Data/PracticeDirectory"))
                                        Console.WriteLine("Make directory operation completed");
                                    else
                                        Console.WriteLine("Make directory operation Failed");
                                    break;
                                case 5:
                                    Console.WriteLine("set replication 5 for hdfs location " +
                                                      "'/Data/PracticeTemp.txt'");
                                    if (p.SetReplication(hostname, clusterType, "/Data/PracticeTemp.txt", 5))
                                        Console.WriteLine("Set replication operation completed");
                                    else
                                        Console.WriteLine("Set Replication operation Failed");
                                    break;
                                case 6:
                                    Console.WriteLine("Enter the offset value:");
                                    string offset=Console.ReadLine();
                                      Console.WriteLine("Uploading file from '" + commonHome +
                                                      "\\BigDataSDK\\Samples\\Data\\WarPeace.txt'" +
                                                      " to hdfs location " + "'/Data/PracticeTempOffset.txt'");
                                    p.UploadFile(hostname, clusterType,
                                        commonHome + "\\BigDataSDK\\Samples\\Data\\WarPeace.txt",
                                        "/Data/PracticeTempOffset.txt", Convert.ToInt32(offset));
                                    break;
                                case 7:
                                    break;
                            }
                        } while (i != 7);
                        if (i == 7)
                        {
                            Console.WriteLine(
                                "Press 'Y' to continue by connecting new cluster or Press any key to exit:");
                            exitOption = Console.ReadLine();
                            if (exitOption.ToUpper() != "Y")
                                break;
                        }
                    }
                    else
                    {
                        Console.WriteLine("Invalid Choice\nPress 'Y' to try again or Press any key to exit:");
                        clusterChoice = (Console.ReadLine().ToUpper() == "Y");
                    }
                } while (clusterChoice);
            }
            catch (Exception ex)
            {
                //TODO
            }
        }

        /// <summary>
        /// Upload file operation
        /// </summary>
        /// <param name="host"></param>
        /// <param name="clusterType"></param>
        /// <param name="fileToUpload"></param>
        /// <param name="hdfspathToUpload"></param>
        private void UploadFile(string host, ClusterType clusterType,string fileToUpload, string hdfspathToUpload,int offset=0)
        {
            try
            {
                string url = string.Empty;
                hdfspathToUpload = hdfspathToUpload.TrimEnd('/').TrimStart('/');

                //Creating the url
                url = (clusterType.Equals(ClusterType.Secure))
                    ? ("https://" + host + ":50470/webhdfs/v1/" + hdfspathToUpload + "/?op=CREATE&data=true").Replace(
                        "#", "%23")
                    : (clusterType.Equals(ClusterType.Azure))
                        ? ("https://" + host + ":8000/webhdfs/v1/" + hdfspathToUpload + "/?user.name=" + userName +
                           "&op=CREATE&data=true").Replace("#", "%23").Replace("+", "%2b")
                        : ("http://" + host + ":50070/webhdfs/v1/" + hdfspathToUpload +
                           "/?user.name=SYSTEM&op=CREATE&data=true").Replace("#", "%23");

                var webRequest = WebRequest.Create(url);
                if (clusterType.Equals(ClusterType.Secure))
                    webRequest.Credentials = new NetworkCredential(userName, password);
                else if (clusterType.Equals(ClusterType.Azure))
                {
                    webRequest.Headers.Add("Username", userName);
                    webRequest.Headers.Add("Password", password);
                }
                webRequest.Method = "PUT";
                const int chunkSize = 1024 * 1024; // read the file by chunks of 10 MB
                using (var file = new FileStream(fileToUpload, FileMode.Open, FileAccess.Read))
                {
                    file.Seek(offset, SeekOrigin.Begin);
                    var buffer = new byte[file.Length- offset];
                    long fileLength = file.Length- offset;
                    if (fileLength <= chunkSize)
                        buffer = new byte[file.Length- offset];
                   

                    webRequest.ContentType = "application/octet-stream";
                    webRequest.ContentLength = buffer.Length;
                    file.Read(buffer, 0, (int)fileLength);
                    Stream objStream = webRequest.GetRequestStream();
                    objStream.Write(buffer, 0, (int)fileLength);

                    objStream.Close();
                    objStream = webRequest.GetResponse().GetResponseStream();
                    url = (clusterType.Equals(ClusterType.Secure))
                        ? ("https://" + host + ":50470/webhdfs/v1/" + hdfspathToUpload +
                           "/?op=SETREPLICATION&replication=1").Replace("#", "%23")
                        : (clusterType.Equals(ClusterType.Azure))
                            ? ("https://" + host + ":8000/webhdfs/v1/" + hdfspathToUpload + "/?user.name=SYSTEM" +
                               "&op=SETREPLICATION&replication=1").Replace("#", "%23")
                            : ("http://" + host + ":50070/webhdfs/v1/" + hdfspathToUpload +
                               "/?user.name=SYSTEM" + "&op=SETREPLICATION&replication=1").Replace("#", "%23");
                    SendWebRequest(url, "PUT",clusterType);

                    buffer = (file.Length - file.Position) > chunkSize ? new byte[chunkSize] : new byte[(file.Length - file.Position)];
                    while (file.Read(buffer, 0, buffer.Length) > 0)
                    {
                        //Append File
                        url = (clusterType.Equals(ClusterType.Secure))
                            ? ("https://" + host + ":50470/webhdfs/v1/" +
                               hdfspathToUpload +
                               "/?Op=APPEND&buffersize=" + webRequest.ContentLength)
                            : (clusterType.Equals(ClusterType.Azure))
                                ? ("https://" + host + ":8000/webhdfs/v1/" + hdfspathToUpload + "/?user.name=" +
                                   userName + "&op=APPEND&data=true")
                                : ("http://" + host + ":50070/webhdfs/v1/" +
                                   hdfspathToUpload +
                                   "/?user.name=SYSTEM&Op=APPEND&buffersize=" + webRequest.ContentLength);
                        WebRequest webRequest1 = WebRequest.Create(url);
                        webRequest1.Credentials = new NetworkCredential(userName, password);
                        webRequest1.Method = "POST";
                        webRequest1.ContentType = "application/octet-stream";
                        webRequest1.ContentLength = buffer.Length;
                        Stream objStreamAppend = webRequest1.GetRequestStream();
                        objStreamAppend.Write(buffer, 0, buffer.Length);
                        objStreamAppend.Close();
                        webRequest1.GetResponse().GetResponseStream();
                        buffer = (file.Length - file.Position) > chunkSize ? new byte[chunkSize] : new byte[(file.Length - file.Position)];
                    }
                }
                if (host == "localhost" || host == "127.0.0.1")
                    url = ("http://" + host + ":50070/webhdfs/v1/" + hdfspathToUpload +
                         "/?user.name=SYSTEM" + "&op=SETREPLICATION&replication=1").Replace("#", "%23");
                else
                {
                    url = (clusterType.Equals(ClusterType.Secure))
                        ? ("https://" + host + ":50470/webhdfs/v1/" + hdfspathToUpload +
                           "/?op=SETREPLICATION&replication=3").Replace("#", "%23")
                        : (clusterType.Equals(ClusterType.Azure))
                            ? ("https://" + host + ":8000/webhdfs/v1/" + hdfspathToUpload + "/?user.name=SYSTEM" +
                               "&op=SETREPLICATION&replication=3").Replace("#", "%23")
                            : ("http://" + host + ":50070/webhdfs/v1/" + hdfspathToUpload +
                               "/?user.name=SYSTEM" + "&op=SETREPLICATION&replication=3").Replace("#", "%23");
                }
                SendWebRequest(url, "PUT",clusterType);
            }
            catch (Exception e)
            {
                Console.Error.WriteLine(
                    "/**UploadHdfsFile(string nameNodeIP, string fileToUpload, string hdfspathToUpload)**/" +
                    e.Message + "\n" + e.StackTrace);
            }
        }

        /// <summary>
        /// Deletes a selected file
        /// </summary>
        /// <param name="clusterType"></param>
        /// <param name="filePath">HDFS path of file needs to be deleted</param>
        /// <param name="host"></param>
        /// <returns>If Delete succeeds returns true else returns false</returns>
        public bool DeleteFile(string host, ClusterType clusterType, string filePath)
        {
            string url = (clusterType.Equals(ClusterType.Secure))
                ? ("https://" + host + ":50470/webhdfs/v1/" + filePath.TrimStart('/') + "/?op=DELETE&recursive=true")
                    .Replace("#", "%23")
                : (clusterType.Equals(ClusterType.Azure))
                    ? ("https://" + host + ":8000/webhdfs/v1/" + filePath.TrimStart('/') + "/?user.name=" + userName +
                       "&op=DELETE&recursive=true").Replace("#", "%23").Replace("+", "%2b")
                    : ("http://" + host + ":50070/webhdfs/v1/" + filePath.TrimStart('/') +
                       "/?user.name=SYSTEM&op=DELETE&recursive=true").Replace("#", "%23").Replace("+", "%2b");
            return SendWebRequest(url, "DELETE", clusterType);
        }

        /// <summary>
        /// Creates a new folder
        /// </summary>
        /// <param name="clusterType"></param>
        /// <param name="folderPath">HDFS Path of folder to be created</param>
        /// <param name="host"></param>
        /// <returns>If folder is created returns true else returns false</returns>
        public bool CreateDirectory(string host, ClusterType clusterType, string folderPath)
        {
            string url = (clusterType.Equals(ClusterType.Secure))
                ? ("https://" + host + ":50470/webhdfs/v1/" + folderPath.TrimStart('/') + "/?op=MKDIRS").Replace("#",
                    "%23")
                : (clusterType.Equals(ClusterType.Azure))
                    ? ("https://" + host + ":8000/webhdfs/v1/" + folderPath.TrimStart('/') + "/?user.name=" + userName +
                       "&op=MKDIRS").Replace("#", "%23").Replace("+", "%2b")
                    : ("http://" + host + ":50070/webhdfs/v1/" + folderPath.TrimStart('/') +
                       "/?user.name=SYSTEM&op=MKDIRS")
                        .Replace("#", "%23");
            return SendWebRequest(url, "PUT",clusterType);

        }

        /// <summary>
        ///  Renames Selected file/folder(s)
        /// </summary>
        /// <param name="host"></param>
        /// <param name="clusterType"></param>
        /// <param name="sourceName">Current file name</param>
        /// <param name="targetName">New name for the file</param>
        /// <returns>If Rename succeeds returns true else returns false</returns>
        public bool MoveFile(string host, ClusterType clusterType,string sourceName, string targetName)
        {
            string commonHome = Directory.GetParent("..\\..\\..\\..\\..\\..\\..\\..\\").ToString();
            UploadFile(host, clusterType, commonHome + "\\BigDataSDK\\Samples\\Data\\WarPeace.txt", targetName);
            DeleteFile(host, clusterType, "/Data/PracticeTemp.txt");
            return true;
        }

        /// <summary>
        ///  Renames Selected file/folder(s)
        /// </summary>
        /// <param name="host"></param>
        /// <param name="clusterType"></param>
        /// <param name="sourceName">Current file name</param>
        /// <param name="replication"></param>
        /// <returns>If Rename succeeds returns true else returns false</returns>
        public bool SetReplication(string host, ClusterType clusterType, string sourceName, int replication)
        {
            string url = (clusterType.Equals(ClusterType.Secure))
                ? ("https://" + host + ":50470/webhdfs/v1" + sourceName + "/?op=SETREPLICATION&replication=" +
                   replication).Replace("#", "%23")
                : (clusterType.Equals(ClusterType.Azure))
                    ? ("https://" + host + ":8000/webhdfs/v1/" + sourceName + "/?user.name=SYSTEM" +
                       "&op=SETREPLICATION&replication=" + replication).Replace("#", "%23")
                    : ("http://" + host + ":50070/webhdfs/v1" + sourceName +
                       "/?user.name=SYSTEM&op=SETREPLICATION&replication=" + replication).Replace("#", "%23");
            return SendWebRequest(url, "PUT", clusterType);
        }

        /// <summary>
        /// Send webrequest for given URL and protocol
        /// </summary>
        /// <param name="url">URL</param>
        /// <param name="protocol">Protocol</param>
        /// <param name="clusterType"></param>
        /// <returns>Returns success status of web request <true>return true if success</true></returns>
        private bool SendWebRequest(string url, string protocol, ClusterType clusterType)
        {
            try
            {
                WebRequest request = WebRequest.Create(url);
                if (clusterType.Equals(ClusterType.Secure))
                    request.Credentials = new NetworkCredential(userName, password);
                else if (clusterType.Equals(ClusterType.Azure))
                {
                    request.Headers.Add("Username", userName);
                    request.Headers.Add("Password", password);
                }
                request.Method = protocol;
                using (var response = request.GetResponse())
                {
                    response.Close();
                }
                return true;
            }
            catch (Exception e)
            {
                Console.Error.WriteLine("/**SendWebRequest(string url, string protocol)**/" + e.Message + "\n" + e.StackTrace);
                return false;
            }
        }

        /// <summary>
        /// Certificate validation
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="certificate"></param>
        /// <param name="chain"></param>
        /// <param name="sslPolicyErrors"></param>
        /// <returns></returns>
        public static bool ValidateServerCertificate(object sender, X509Certificate certificate,
            X509Chain chain,
            SslPolicyErrors sslPolicyErrors)
        {
            return true;
        }

        /// <summary>
        /// Used to get host name
        /// </summary>
        /// <param name="ip"></param>
        /// <returns></returns>
        public static string GetHostName(string ip)
        {
            try
            {
                var hostEntry = Dns.GetHostByAddress(ip);
                return hostEntry.HostName;
            }
            catch(Exception ex)
            {
                return ip;
            }
        }

    }

    public enum ClusterType
    {
        Normal,
        Secure,
        Azure
    }
}
